/*
 * DB: View Editor (edits the views of a DataSource)
 * Copyright (C) 1998-2000  by Tom Dyas (tdyas@vger.rutgers.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>

#include "enum.h"
#include "db.h"
#include "grid.h"

#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

#define fieldNameColumn  0
#define fieldWidthColumn 1

typedef struct {
    UInt16 fieldNum;
    UInt16 width;
    UInt32 uniqueID;
} ColInfo;

static UInt16 numCols = 0;
static UInt32 NextUniqueID;
static ColInfo * cols = 0;
static MemHandle cols_handle = 0;
static MemHandle * width_handles = 0;
static MemHandle width_handles_handle = 0;
static char ** field_names = 0;

static UInt16 TopVisibleField;
static MemHandle name_handle;

static Char name_trigger_label[32];

static void UpdateScrollers(void) SECT_UI;
static Err LoadData(void *, Int16, Int16, Boolean, MemHandle *, Int16 *,
		    Int16 *, FieldPtr) SECT_UI;
static void LoadTable(TablePtr table) SECT_UI;
static void InitTable(TablePtr table) SECT_UI;
static void Scroll(WinDirectionType direction) SECT_UI;
static Boolean SaveFieldWidths(void) SECT_UI;
static void GetDefinition(UInt16 defNum) SECT_UI;
static void PutDefinition(UInt16 defNum, Boolean doStore) SECT_UI;
static void AddColumn(void) SECT_UI;
static void DeleteColumn(UInt16 colNum) SECT_UI;


static void UpdateScrollers(void)
{
    UInt16 upIndex, downIndex, fieldNum;
    Boolean scrollUp, scrollDown;
    Int16 row;
    FormPtr form;
    TablePtr table;

    scrollUp = TopVisibleField != 0;

    form = FrmGetActiveForm();
    table = GetObjectPtr(form, ctlID_ListViewEditor_Table);
    row = TblGetLastUsableRow(table);
    if (row != -1) {
	fieldNum = TblGetRowID(table, row);
	scrollDown = fieldNum < numCols - 1;
    } else {
	scrollDown = false;
    }

    upIndex = FrmGetObjectIndex(form, ctlID_ListViewEditor_UpButton);
    downIndex = FrmGetObjectIndex(form, ctlID_ListViewEditor_DownButton);
    FrmUpdateScrollers(form, upIndex, downIndex, scrollUp, scrollDown);
}

static Err LoadData(void * table, Int16 row, Int16 column,
		    Boolean editing, MemHandle * textH, Int16 * textOffset,
		    Int16 * textAllocSize, FieldPtr fld)
{
    UInt16 colNum;
    Char * p;
    FieldAttrType attr;
    MemHandle handle;

    colNum = (UInt16) TblGetRowID(table, row);

    *textH = width_handles[colNum];
    *textOffset = 0;
    *textAllocSize = MemHandleSize(width_handles[colNum]);

    if (fld) {
	FldGetAttributes(fld, &attr);
	attr.singleLine = 1;
	attr.dynamicSize = 0;
	attr.underlined = 1;
	FldSetAttributes(fld, &attr);
	
	FldSetMaxChars(fld, 16);
    }

    return (0);
}           

static void LoadTable(TablePtr table)
{
    UInt16 numRows, row, colNum;
    ListPtr list;
    FormPtr form;

    form = FrmGetActiveForm();
    list = GetObjectPtr(form, ctlID_ListViewEditor_PopupList);

    numRows = TblGetNumberOfRows(table);
    colNum = TopVisibleField;
    for (row = 0; row < numRows; row++) {
	if (colNum < numCols) {
	    if (! TblRowUsable(table, row)
		|| TblGetRowData(table, row) != cols[colNum].uniqueID) {
		TblSetRowUsable(table, row, true);
		TblMarkRowInvalid(table, row);

		TblSetItemStyle(table, row,
				fieldNameColumn, popupTriggerTableItem);
		TblSetItemPtr(table, row, fieldNameColumn, list);
		TblSetItemInt(table, row, fieldNameColumn,
			      cols[colNum].fieldNum);

		TblSetItemStyle(table, row, fieldWidthColumn, textTableItem);
		TblSetRowID(table, row, colNum);
		TblSetRowData(table, row, cols[colNum].uniqueID);
	    }
	} else {
	    TblSetRowUsable(table, row, false);
	}
	colNum++;
    }

    UpdateScrollers();
}

static void InitTable(TablePtr table)
{
    UInt16 row, numRows;

    numRows = TblGetNumberOfRows(table);
    for (row = 0; row < numRows; row++)
	TblSetRowUsable(table, row, false);

    TblSetColumnUsable(table, fieldNameColumn, true);
    TblSetColumnWidth(table, fieldNameColumn, 120);

    TblSetColumnUsable(table, fieldWidthColumn, true);
    TblSetColumnWidth(table, fieldWidthColumn, 40);

    TblSetLoadDataProcedure(table, fieldWidthColumn, LoadData);

    LoadTable(table);
}

static void Scroll(WinDirectionType direction)
{
    UInt16 newTopVisibleField;

    newTopVisibleField = TopVisibleField;

    if (direction == winDown) {
	newTopVisibleField++;
    } else {
	newTopVisibleField--;
    }

    if (TopVisibleField != newTopVisibleField) {
	FormPtr form;
	TablePtr table;

	TopVisibleField = newTopVisibleField;

	form = FrmGetActiveForm();
	table = GetObjectPtr(form, ctlID_ListViewEditor_Table);
	LoadTable(table);
	TblRedrawTable(table);
    }
}

static Boolean SaveFieldWidths(void)
{
    UInt16 i;
    UInt16 widths[numCols];
    char *p;

    /* Make sure that each field width is numeric and in range. */
    for (i = 0; i < numCols; ++i) {
	p = MemHandleLock(width_handles[i]);
	if (! IsNumber(p)) {
	    FrmAlert(alertID_BadFieldWidth);
	    MemPtrUnlock(p);
	    return false;
	}
	widths[i] = (UInt16) String2Long(p);
	MemPtrUnlock(p);

	if (widths[i] < 10 || widths[i] > 160) {
	    FrmAlert(alertID_BadFieldWidth);
	    return false;
	}
    }

    /* Move the column widths into the list view definition. */
    for (i = 0; i < numCols; ++i) cols[i].width = widths[i];

    return true;
}

static void
GetDefinition(UInt16 defNum)
{
    DataSourceView * view;
    FormPtr form;
    FieldPtr fld;
    ControlPtr ctl;
    Char * p;
    UInt16 i;

    /* Load the data source view. */
    view = CurrentSource->ops.GetView(CurrentSource, defNum);

    /* Copy the view information into the holding places for editing. */
    numCols = view->numCols;
    cols_handle = MemHandleNew(numCols * sizeof(ColInfo));
    width_handles_handle = MemHandleNew(numCols * sizeof(MemHandle));
    cols = MemHandleLock(cols_handle);
    width_handles = MemHandleLock(width_handles_handle);
    NextUniqueID = 0;
    for (i = 0; i < numCols; ++i) {
	cols[i].fieldNum = view->cols[i].fieldNum;
	cols[i].width = view->cols[i].width;
	cols[i].uniqueID = NextUniqueID++;
	width_handles[i] = MemHandleNew(16);
	p = MemHandleLock(width_handles[i]);
	StrPrintF(p, "%u", cols[i].width);
	MemPtrUnlock(p);
    }

    /* Disconnect name_handle from the name field. */
    form = FrmGetActiveForm();
    fld = GetObjectPtr(form, ctlID_ListViewEditor_NameField);
    FldSetTextHandle(fld, 0);

    /* Store the name of this definition in the handle for the name. */
    p = MemHandleLock(name_handle);
    StrCopy(p, view->name);
    MemPtrUnlock(p);

    /* Reconnect the name field to name_handle. */
    FldSetTextHandle(fld, name_handle);

    /* Copy the name to the trigger label. */
    StrCopy(name_trigger_label, view->name);
    ctl = GetObjectPtr(form, ctlID_ListViewEditor_NameTrigger);
    CtlSetLabel(ctl, name_trigger_label);

    /* Free the data source view since we have copied it to local storage. */
    CurrentSource->ops.PutView(CurrentSource, defNum, view);
}

static void
PutDefinition(UInt16 defNum, Boolean doStore)
{
    DataSourceView view;
    UInt16 i;
    Char * p, *q;

    if (doStore) {
	/* Copy the name field back into the view. */
	p = MemHandleLock(name_handle);
	StrNCopy(view.name, p, sizeof(view.name) - 1);
	view.name[sizeof(view.name)-1] = '\0';
	MemPtrUnlock(p);

	/* Construct the column information. */
	view.numCols = numCols;
	view.cols = MemPtrNew(numCols * sizeof(DataSourceViewColumn));
	for (i = 0; i < numCols; ++i) {
	    view.cols[i].fieldNum = cols[i].fieldNum;
	    view.cols[i].width = cols[i].width;
	    MemHandleFree(width_handles[i]);
	}

	/* Tell the data source to update this view. */
	CurrentSource->ops.StoreView(CurrentSource, defNum, &view);

	/* Free the pointer we allocated before. */
	MemPtrFree(view.cols);
    }

    /* Free the global structures used for this form's state. */
    MemPtrUnlock(cols); cols = 0;
    MemHandleFree(cols_handle); cols_handle = 0;
    MemPtrUnlock(width_handles); width_handles = 0;
    MemHandleFree(width_handles_handle); width_handles_handle = 0;
}

static void
AddColumn(void)
{
    char * p;

    /* Reallocate the cols_handle to fit new column. */
    MemHandleUnlock(cols_handle);
    MemHandleResize(cols_handle, (numCols + 1) * sizeof(ColInfo));
    cols = MemHandleLock(cols_handle);

    /* Fill in the new column. */
    cols[numCols].fieldNum = 0;
    cols[numCols].width = 80;
    cols[numCols].uniqueID = NextUniqueID++;

    /* Reallocate the width_handles_handle to fit new column. */
    MemHandleUnlock(width_handles_handle);
    MemHandleResize(width_handles_handle, (numCols + 1) * sizeof(MemHandle));
    width_handles = MemHandleLock(width_handles_handle);

    /* Fill in the new column. */
    width_handles[numCols] = MemHandleNew(16);
    p = MemHandleLock(width_handles[numCols]);
    StrPrintF(p, "%u", cols[numCols].width);
    MemPtrUnlock(p);

    /* Increment the number of columns. */
    numCols++;

    /* If the number of columns was 1, enable the remove option. */
    if (numCols > 1 && ! CurrentSource->readonly) {
	FormPtr form;
	ListPtr list;

	form = FrmGetActiveForm();
	list = GetObjectPtr(form, ctlID_ListViewEditor_PopupList);
	LstSetListChoices(list, field_names,
			  1 + CurrentSource->schema.numFields);
	LstSetHeight(list, 1 + CurrentSource->schema.numFields);
    }
}

static void
DeleteColumn(UInt16 colNum)
{
    UInt16 i;

    /* Free the width handle associated with the named column. */
    MemHandleFree(width_handles[colNum]);

    /* Move every width handle and column info above this point down. */
    for (i = colNum + 1; i < numCols; ++i) {
	cols[i-1] = cols[i];
	width_handles[i-1] = width_handles[i];
    }

    /* Reallocate the cols_handle to fit. */
    MemHandleUnlock(cols_handle);
    MemHandleResize(cols_handle, (numCols - 1) * sizeof(ColInfo));
    cols = MemHandleLock(cols_handle);

    /* Reallocate the width_handles_handle to fit. */
    MemHandleUnlock(width_handles_handle);
    MemHandleResize(width_handles_handle, (numCols - 1) * sizeof(MemHandle));
    width_handles = MemHandleLock(width_handles_handle);

    /* Decrement the number of columns. */
    numCols--;

    /* If the number of columns goes to 1, disable the remove option. */
    if (numCols == 1) {
	FormPtr form;
	ListPtr list;

	form = FrmGetActiveForm();
	list = GetObjectPtr(form, ctlID_ListViewEditor_PopupList);
	LstSetListChoices(list, field_names, CurrentSource->schema.numFields);
	LstSetHeight(list, CurrentSource->schema.numFields);
    }
}

Boolean
ListViewEditor_HandleEvent(EventPtr event)
{
    FormPtr form;
    TablePtr table;
    ListPtr list;
    FieldPtr fld;
    UInt16 i;
    Char * p;

    form = FrmGetActiveForm();

    switch (event->eType) {
    case frmOpenEvent:
	/* Allocate a handle for the definition name. */
	name_handle = MemHandleNew(32);

	/* Load the definition that will be edited. */
	GetDefinition(0);

	/* Allocate and fill the field names for the popup list. */
	field_names = MemPtrNew((CurrentSource->schema.numFields + 1)
				* sizeof(char *));
	for (i = 0; i < CurrentSource->schema.numFields; ++i) {
	    field_names[i] = CurrentSource->schema.fields[i].name;
	}
	p = GetStringPtr(stringID_Remove);
	field_names[CurrentSource->schema.numFields] = MemPtrNew(StrLen(p)+1);
	StrCopy(field_names[CurrentSource->schema.numFields], p);
	MemPtrUnlock(p);

	/* Make the popup list use the field names data. */
	list = GetObjectPtr(form, ctlID_ListViewEditor_PopupList);
	if (numCols > 1 && ! CurrentSource->readonly) {
	    LstSetListChoices(list, field_names, 1 +
			      CurrentSource->schema.numFields);
	    LstSetHeight(list, 1 + CurrentSource->schema.numFields);
	} else {
	    LstSetListChoices(list, field_names,
			      CurrentSource->schema.numFields);
	    LstSetHeight(list, CurrentSource->schema.numFields);
	}

	/* Draw the form. */
	TopVisibleField = 0;
	table = GetObjectPtr(form, ctlID_ListViewEditor_Table);
	InitTable(table);
	FrmDrawForm(form);

	return true;

    case frmCloseEvent:
        /* Ensure that the table no longer has focus. */
	table = GetObjectPtr(form, ctlID_ListViewEditor_Table);
	if (TblGetCurrentField(table)) {
	    TblReleaseFocus(table);
	}

	/* Store and free any state associated with local copy. */
	if (cols) {
	    PutDefinition(0, false);
	}

	/* Free the space allocated for the field_names array. */
	MemPtrFree(field_names[CurrentSource->schema.numFields]);
	MemPtrFree(field_names);
	field_names = 0;

	break;

    case tblSelectEvent:
	if (event->data.tblSelect.column == fieldNameColumn) {
	    UInt16 colNum = TblGetRowID(event->data.tblSelect.pTable,
					event->data.tblSelect.row);
	    UInt16 fieldNum = TblGetItemInt(event->data.tblSelect.pTable,
					    event->data.tblSelect.row,
					    event->data.tblSelect.column);
	    if (fieldNum < CurrentSource->schema.numFields)
	      cols[colNum].fieldNum = fieldNum;
	    else {
		/* User picked the remove option. */
		DeleteColumn(colNum);
		table = GetObjectPtr(form, ctlID_ListViewEditor_Table);
		LoadTable(table);
		TblRedrawTable(table);
	    }
	}
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case ctlID_ListViewEditor_DoneButton:
	    /* If the data source is in read-only mode, discard any edits. */
	    if (CurrentSource->readonly) {
		PutDefinition(0, false);
		FrmGotoForm(formID_ListView);
		return true;
	    }
	    if (SaveFieldWidths()) {
		PutDefinition(0, true);
		FrmGotoForm(formID_ListView);
	    }
	    return true;

	case ctlID_ListViewEditor_CancelButton:
	    PutDefinition(0, false);
	    FrmGotoForm(formID_ListView);
	    return true;

	case ctlID_ListViewEditor_AddButton:
	    AddColumn();
	    table = GetObjectPtr(form, ctlID_ListViewEditor_Table);
	    LoadTable(table);
	    TblRedrawTable(table);
	    return true;
	}
	break;

    case ctlRepeatEvent:
	switch (event->data.ctlRepeat.controlID) {
	case ctlID_ListViewEditor_UpButton:
	    Scroll(winUp);
	    break;

	case ctlID_ListViewEditor_DownButton:
	    Scroll(winDown);
	    break;
	}
	break;

    case menuEvent:
	return HandleCommonMenuEvent(event->data.menu.itemID);

    }

    return false;
}
